@@ -1,5 +1,4 @@
 require 'prb/prb'
-require 'prb/opt'
 
 class World
   class UserList < PRb::PRbRoot
@@ -28,7 +27,7 @@ class World
       world = root[$0]
       unless world
 	world = UserList.knew
-	$prb.root[$0] = world
+	root[$0] = world
       end
       @world = world
     end
@@ -56,8 +55,7 @@ class World
 end
 
 def main
-  pool = DBIPool.new(*DBIPool.check_options(ARGV))
-  $prb = PRb::Store.new(pool)
+  PRb.start_service(ARGV)
   world = World.new
   world.demo
 end
