@@ -1,4 +1,16 @@
-require 'prb/dbiu'
+require 'dbi'
+require 'dbiu/dbipool'
+require 'thread'
+
+module DBI
+  module SQL
+    def SQL.query?(sql)
+      return false unless /^\s*select\b/i =~ sql
+      return false if /\s+for\s+update/im =~ sql
+      true
+    end
+  end
+end
 
 module PRb
   MAX_INT8 = 2**63-1
@@ -14,7 +26,7 @@ module PRb
     end
 
     def inspect
-      "#<#{type} ref=#{@ref}>"
+      "#<#{self.class} ref=#{@ref}>"
     end
 
     def transaction
@@ -173,7 +185,7 @@ module PRb
     end
 
     def ==(other)
-      return false unless type === other
+      return false unless self.class === other
       @ref == other._prb_ref && @prb.db_name == other._prb_db_name
     end
   end
@@ -316,13 +328,23 @@ module PRb
       @name = name
     end
 
-    def next_value
+    def next_value(n=1)
       @pool.transaction(true) do
-	conn.do("update seq set value=value+incr where name=?", @name)
-	conn.select_one("select value from seq where name=?", @name)[0]
+	conn.do("update seq set value=value+incr*? where name=?", n, @name)
+	incr, val = conn.select_one("select incr, value from seq where name=?",
+				    @name)
+	if n == 1
+	  val
+	else
+	  ary = Array.new(n)
+	  n.times do |i|
+	    ary[i] = val - incr * (n - i - 1)
+	  end
+	  ary
+	end
       end
     end
-    
+
     def create(start, incr)
       start = start - incr
       @pool.transaction(true) do
@@ -331,6 +353,34 @@ module PRb
     end
   end
 
+  class SeqQueue
+    def initialize(pool, name)
+      @seq = Seq.new(pool, name)
+      @queue = SizedQueue.new(4)
+      @thread = nil
+    end
+    
+    def next_value
+      start_thread unless @thread
+      @queue.pop
+    end
+
+    def start_thread
+      @thread = Thread.new do
+	while true
+	  ary = @seq.next_value(8)
+	  ary.each do |x|
+	    @queue.push(x)
+	  end
+	end
+      end
+    end
+
+    def create(start, incr)
+      @seq.create(start, incr)
+    end
+  end
+
   class SymbolTable
     include DBIPool::DBIPoolConn
     def initialize(pool)
@@ -446,13 +496,14 @@ module PRb
       @symcache = {}
       
       @symtab = SymbolTable.new(pool)
-      @prbseq = Seq.new(pool, 'object_seq')
+      # @prbseq = Seq.new(pool, 'object_seq')
+      @prbseq = SeqQueue.new(pool, 'object_seq')
       
       @root_id = 8
       @root = PRbRoot.new(self, @root_id)
       
       @db_name = @pool.db_name
-      type[@db_name] = self
+      self.class[@db_name] = self
     end
     attr_reader :root, :root_id, :db_name
     
@@ -482,7 +533,13 @@ create table alist (id int8 ,
 EOQ
       conn.do(sql)
 
-      conn.do("select * into o_gc from object")
+      sql = <<EOQ
+create table o_gc (id int8, 
+                   klass int8, value text, mark boolean,
+		   primary key (id));
+EOQ
+      conn.do(sql)
+
       conn.do("select * into a_gc from alist")
     end
 
@@ -725,34 +782,43 @@ EOQ
 	conn.do("select * from object for update")
 	conn.do("select * from alist for update")
 
-	conn.do("delete from a_gc")
 	conn.do("delete from o_gc")
+	conn.do("insert into o_gc select object.*, false from object")
 	
-	conn.do("insert into o_gc select * from object where id=?", @root_id)
-	conn.do("insert into a_gc select * from alist where id=?", @root_id)
-	
+	conn.do("update o_gc set mark=true where id=?", @root_id)
 	loop do
-	  sql =<<EOQ
-from object as o, a_gc 
-   where o.id not in (select o_gc.id from o_gc) 
-    and (a_gc.key = o.id or a_gc.value = o.id);
+	  sql = <<EOQ
+select count(*) from o_gc as o, o_gc as gc, alist as a
+where o.mark=false and o.id=a.value and a.id=gc.id and gc.mark=true
 EOQ
-	  count, = conn.select_one("select count(*) #{sql}")
+	  count, = conn.select_one(sql)
 	  break if count == 0
 	  
-	  conn.do("insert into o_gc select distinct o.* #{sql}")
 	  sql = <<EOQ
-insert into a_gc select a.* from alist as a, o_gc as o
-  where a.id not in (select id from a_gc) and o.id=a.id
+update o_gc set mark=true from o_gc as gc, alist as a
+where o_gc.mark=false and o_gc.id=a.value and a.id=gc.id and gc.mark=true;
 EOQ
 	  conn.do(sql)
 	end
 
-	conn.do("delete from alist")
-	conn.do("delete from object")
+	conn.do('delete from a_gc')
+	
+	sql = <<EOQ
+insert into a_gc select a.* from alist as a, o_gc as o
+where a.id=o.id and o.mark=true
+EOQ
+	conn.do(sql)
+
+	conn.do('delete from alist')
+	conn.do('delete from object')
+
+	sql = <<SQL
+insert into object select o_gc.id, o_gc.klass, o_gc.value
+from o_gc where o_gc.mark=true
+SQL
+	conn.do(sql)
+	conn.do('insert into alist select * from a_gc')
 	
-	conn.do("insert into object select * from o_gc")
-	conn.do("insert into alist select * from a_gc")
       end
     end
 
@@ -774,9 +840,9 @@ EOQ
     end
     
     def prop_to_obj(ref, klass, value)
-      return true if ref == true.id
-      return false if ref == false.id
-      return nil if ref == nil.id
+      return true if ref == true.__id__
+      return false if ref == false.__id__
+      return nil if ref == nil.__id__
       return @symtab[ref] if (ref - 2) % 4 == 0
       return (ref - 1) / 2 if (ref - 1) % 2 == 0
       
@@ -799,7 +865,7 @@ EOQ
     def x_obj2ref(obj)
       case obj
       when true, false, nil, Fixnum
-	obj.id
+	obj.__id__
       when Symbol
 	@symtab.symbol(obj)
       when PRbO
@@ -812,10 +878,10 @@ EOQ
     def ref_walk(obj, map)
       ref = x_obj2ref(obj)
       return ref unless ref.nil?
-      return map[obj.id] if map.include? obj.id
+      return map[obj.__id__] if map.include? obj.__id__
       
-      ref = alloc_obj(obj.type, prim_dump(obj))
-      map[obj.id] = ref
+      ref = alloc_obj(obj.class, prim_dump(obj))
+      map[obj.__id__] = ref
       
       case obj
       when Array
@@ -839,4 +905,10 @@ EOQ
     Store.primary
   end
   module_function :primary
+
+  def start_service(argv=[], opt={ :db=>'DBI:Pg:prb' })
+    pool = DBIPool.new(*DBIPool.check_options(argv, opt))
+    PRb::Store.new(pool)
+  end
+  module_function :start_service
 end
